/*
 * Copyright (C) 2009-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file oss_write.c Paraslash's oss output plugin. */

#include <regex.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/soundcard.h>

#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "oss_write.cmdline.h"
#include "error.h"

/** Data specific to the oss writer. */
struct private_oss_write_data {
	/** The file handle of the device. */
	int fd;
	/** Four bytes for stereo streams, two bytes for mono streams. */
	int bytes_per_frame;
};

static int get_oss_format(enum sample_format sf)
{
	switch (sf) {
	case SF_S8: return AFMT_S8;
	case SF_U8: return AFMT_U8;
	case SF_S16_LE: return AFMT_S16_LE;
	case SF_S16_BE: return AFMT_S16_BE;
	case SF_U16_LE: return AFMT_U16_LE;
	case SF_U16_BE: return AFMT_U16_BE;
	default: return AFMT_S16_LE;
	}
}

static void oss_pre_select(struct sched *s, struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_oss_write_data *powd = wn->private_data;
	int ret = btr_node_status(wn->btrn, wn->min_iqs, BTR_NT_LEAF);

	if (ret == 0)
		return;
	if (ret < 0 || !powd)
		return sched_min_delay(s);
	para_fd_set(powd->fd, &s->wfds, &s->max_fileno);
}

static void oss_close(struct writer_node *wn)
{
	struct private_oss_write_data *powd = wn->private_data;

	if (!powd)
		return;
	close(powd->fd);
	free(powd);
}

/*
 * The Open Sound System Programmer's Guide sayeth:
 *
 * Set sampling parameters always so that number of channels (mono/stereo) is
 * set before selecting sampling rate (speed).  Failing to do this will make
 * your program incompatible with cards such as the SoundBlaster Pro which
 * supports 44.1 kHz in mono but just 22.05 kHz in stereo. A program which
 * selects 44.1 kHz speed and then sets the device to stereo mode will
 * incorrectly believe that the device is still in 44.1 kHz mode when actually
 * the speed is decreased to 22.05 kHz.
 */
static int oss_init(struct writer_node *wn, unsigned sample_rate,
		unsigned channels, int sample_format)
{
	int ret, format;
	unsigned ch, rate;
	struct oss_write_args_info *conf = wn->conf;
	struct private_oss_write_data *powd = para_calloc(sizeof(*powd));

	PARA_INFO_LOG("opening %s\n", conf->device_arg);
	ret = para_open(conf->device_arg, O_WRONLY, 0);
	if (ret < 0)
		goto err_free;
	powd->fd = ret;
	ret = mark_fd_nonblocking(powd->fd);
	if (ret < 0)
		goto err;
	/* set PCM format */
	sample_format = format = get_oss_format(sample_format);
	ret = ioctl(powd->fd, SNDCTL_DSP_SETFMT, &format);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = -E_BAD_SAMPLE_FORMAT;
	if (format != sample_format)
		goto err;
	/* set number of channels */
	ch = channels;
	ret = ioctl(powd->fd, SNDCTL_DSP_CHANNELS, &ch);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = -E_BAD_CHANNEL_COUNT;
	if (ch != channels)
		goto err;
	if (format == SF_U8 || format == SF_S8)
		powd->bytes_per_frame = ch;
	else
		powd->bytes_per_frame = ch * 2;

	/*
	 * Set sampling rate
	 *
	 * If we request a higher sampling rate than is supported by the
	 * device, the the highest possible speed is automatically used. The
	 * value actually used is returned as the new value of the argument.
	 */
	rate = sample_rate;
	ret = ioctl(powd->fd, SNDCTL_DSP_SPEED, &rate);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	if (rate != sample_rate) {
		unsigned min = PARA_MIN(rate, sample_rate),
			max = PARA_MAX(rate, sample_rate);
		/*
		 * Check whether the returned sample rate differs significantly
		 * from the requested one.
		 */
		ret = -E_BAD_SAMPLERATE;
		if (100 * max > 110 * min) /* more than 10% deviation */
			goto err;
		PARA_NOTICE_LOG("using %dHz rather than %dHz\n", rate,
			sample_rate);
	}
	wn->min_iqs = powd->bytes_per_frame;
	wn->private_data = powd;
	return 1;
err:
	close(powd->fd);
err_free:
	free(powd);
	PARA_ERROR_LOG("failed to init %s: %s\n", conf->device_arg,
		para_strerror(-ret));
	return ret;
}

static void oss_post_select(__a_unused struct sched *s,
		struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_oss_write_data *powd = wn->private_data;
	struct btr_node *btrn = wn->btrn;
	size_t frames, bytes;
	int ret = btr_node_status(btrn, wn->min_iqs, BTR_NT_LEAF);
	char *data;

	if (ret < 0)
		goto out;
	if (ret == 0)
		return;
	if (!powd) {
		int32_t rate, ch, format;
		get_btr_sample_rate(btrn, &rate);
		get_btr_channels(btrn, &ch);
		get_btr_sample_format(btrn, &format);
		ret = oss_init(wn, rate, ch, format);
		if (ret < 0)
			goto out;
		return;
	}
	btr_merge(btrn, wn->min_iqs);
	bytes = btr_next_buffer(btrn, &data);
	frames = bytes / powd->bytes_per_frame;
	if (!frames) { /* eof and less than a single frame available */
		ret = -E_WRITE_COMMON_EOF;
		goto out;
	}
	ret = 0;
	if (!FD_ISSET(powd->fd, &s->wfds))
		goto out;
	ret = xwrite(powd->fd, data, frames * powd->bytes_per_frame);
	if (ret < 0)
		goto out;
	btr_consume(btrn, ret);
	ret = 0;
out:
	t->error = ret;
	if (ret < 0)
		btr_remove_node(&wn->btrn);
}

__malloc static void *oss_parse_config_or_die(const char *options)
{
	struct oss_write_args_info *conf = para_calloc(sizeof(*conf));

	/* exits on errors */
	oss_cmdline_parser_string(options, conf, "oss_write");
	return conf;
}

static void oss_free_config(void *conf)
{
	oss_cmdline_parser_free(conf);
}

/**
 * The init function of the oss writer.
 *
 * \param w Pointer to the writer to initialize.
 *
 * \sa struct writer.
 */
void oss_write_init(struct writer *w)
{
	struct oss_write_args_info dummy;

	oss_cmdline_parser_init(&dummy);
	w->close = oss_close;
	w->pre_select = oss_pre_select;
	w->post_select = oss_post_select;
	w->parse_config_or_die = oss_parse_config_or_die;
	w->free_config = oss_free_config;
	w->help = (struct ggo_help) {
		.short_help = oss_write_args_info_help,
		.detailed_help = oss_write_args_info_detailed_help
	};
	oss_cmdline_parser_free(&dummy);
}
